
// bibhtml: Generates HTML page (with images) from BibTex files
// See: https://www.telecom-paris.fr/~elc/software/
//
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <iostream>
#include <fstream>
#include <map>
#include <string>
#include "bib.hpp"
#include "bibentry.hpp"
#include "bibstyle.hpp"
#include "bibtex.hpp"
#include "bibstats.hpp"
#include "extension.hpp"
using namespace std;

Bib Bib::instance;

// needed if Hyde is used because {{ and }} then have a special meaning
// #define WITH_HYDE

int main(int argc, char *argv[]) {
  Bib::instance.init(argc, argv);
  return 0;
}

Bib::~Bib() {
  clear();
}

void Bib::init(int argc, char *argv[])
{
  // Note: _style must be created after _bibTypes
  style_ = BibStyle::createStyle(*this, BibSettings::DefaultStyleFormat);
  
  if (!imageDir_.empty() && imageDir_[imageDir_.length()-1]!='/') imageDir_ += '/';

  bool interFound{false}, showFound{false};

  echo("\nStarting BibToHTML:\n");

  // process command line arguments
  for (int k = 1; k < argc; ++k) {
    string arg = argv[k];
    
    if (arg.empty()) continue;
    else if (arg[0]!='-')
      readBib(arg, false);
    else if (arg.size() > 2 && arg[0]=='-' && arg[1]=='-') {
      string cmd, cmdarg;
      ccuty::strsplit(cmd, cmdarg, argv[k]+2, ":");
      doCommand(cmd, cmdarg, interFound, showFound);
    }
    else if (arg.size() > 1 && arg[0]=='-') {
      string cmd, cmdarg;
      ccuty::strsplit(cmd, cmdarg, argv[k]+1, ":");
      doCommand(cmd, cmdarg, interFound, showFound);
    }
  }
  
  // argc > 1 => batch mode, interactive mode otherwise excpet if -inter present
  if (argc > 1) {
    if (!showFound) writeHtml(false);  // print without displaying html page
    echo("");
    if (interFound) interMode_ = true;
    else return;
  }
  
  echo("Type h for help");
  interMode_ = true;

  while (true) {
    cout << "> ";
    string line;
    if (!getline(cin, line)) return;
    if (line.empty()) continue;
    
    // split cmd:arg couples, delimiter is : (or space in interactive mode)
    vector<string> v;
    if (ccuty::strsplit(v, line, ": ", 2) >= 1) {
      doCommand(v[0], (v.size()>1 ? v[1] : ""), interFound, showFound);
    }
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

void Bib::doCommand(const string& command, const string& args,
                    bool& interFound, bool& showFound) {
  if (command.empty()) return;
  string cmd = (command[0]=='-') ?  command.substr(1) : command;
  
  if (cmd == "?") {
    help();
    return;
  }
  
  switch (commands_.names.starts_with(cmd)) {  // cmd can be abridged
    case Commands::read:
      readBib(args, false);
      return;
    case Commands::aux:
      readBib(args, true);
      return;
    case Commands::images:
      imageDir_ = ccuty::trimmed(args);
      if (!imageDir_.empty() && imageDir_[imageDir_.length()-1]!='/') imageDir_ += '/';
      if (interMode_||verboseMode_) cout << "Images in: '" << imageDir_ << "'" <<endl;
       cout << "Images in: '" << imageDir_ << "'" <<endl;
      if (interMode_) writeHtml(true);
      return;
    case Commands::out:
    case Commands::outhtml:
      setHtmlFile(args);
      if (interMode_||verboseMode_) cout << "Produced HTML file: '" << outHtml_ << "'" <<endl;
      return;
    case Commands::outbib:
      setBibFile(args);
      if (interMode_||verboseMode_) cout << "Produced BibText file: '" << outBib_ << "'" <<endl;
      return;
    case Commands::templ:
      readTemplate(args);
      return;
    case Commands::format:
      setFormat(args);
      return;
    case Commands::all:
      select("*");
      return;
    case Commands::select:
      select(args);
      return;
    case Commands::sort:
      sort(args);
      return;
    case Commands::show:
    case Commands::showhtml:
      showFound = true;
      writeHtml(true);
      return;
    case Commands::showbib:
      showFound = true;
      writeBib(true);
      return;
    case Commands::help:
      help();
      return;
    case Commands::i:
    case Commands::inter:
      interFound = true;
      return;
    case Commands::verbose:
      verboseMode_ = (args=="1" || args=="yes" || args=="true");
      return;
    case Commands::clear:
      clear();
      return;
    case Commands::check:
      checkLinks(args);
      return;
    case Commands::stats:
      showStats(args);
      return;
    case Commands::quit:
      exit(0);
    default:
      if (interMode_) echo("!Error: Unknown or ambiguous command (type h for help)");
      else echo("!Error: Unknown or ambiguous command: -"+ cmd + " (-help for help)");
      return;
  }
}


void Bib::echo(const string& msg) {
  cout << msg << endl;
}


void Bib::error(const char* command, const std::string& msg) {
  cout << "! Error: "<< command << ": " << msg << endl;
}


bool Bib::help() {
  echo("Usage: bibhtml bibtex_file [bibtex_file2] ... [options]\n"
       "With:  options: -command1 -command2:argument ...\n"
       "Enclose arguments in double quotes if they contain spaces\n"
       "\n"
       "Enters interactive mode if no argument or if -i[nter] is present\n"
       "In interactive mode:\n"
       " - Only one command per line\n"
       " - The initial - is not needed and : can be replaced by spaces\n"
       "\n"
       "COMMANDS: \n"
       "   -r[ead]:bibfile      : Read BibTex file\n"
       "   -aux:bibfile         : Read auxiliary BibTex file\n"
       "   -t[empl]:htmlfile    : Read HTML template\n"
       "   -ima[ges]:directory  : Set directory containing images\n"
       "   -out[html]:htmlfile  : Set name of the produced HTML file\n"
       "   -outbib:bibfile      : Set name of the produced BibTex file\n"
       "   -f[ormat]:options    : Set printing format and style (format:? for help)\n"
       "\n"
       "   -a[ll]               : Select all entries\n"
       "   -se[lect]:pattern    : Select entries according to pattern (select:? for help)\n"
       "   -so[rt]:pattern      : Sort selected entries according to pattern (sort:? for help)\n"
       "   -show[html]          : Show produced HTML file (opens browser)\n"
       "   -showbib             : Show produced BibTex file (opens browser)\n"
       "   -stats:csvfile       : Show statistics for selected entries\n"
       "\n"
       "   -h[elp]              : Show available commands\n"
       "   -i[nter]             : Set interactive mode\n"
       "   -v[erbose]           : Set verbose mode\n"
       "   -cl[ear]             : Clear the database (remove all entries)\n"
       "   -ch[eck]:field       : Check links for this field (for selected entries)\n"
       "   -q[quit]             : Quit application\n");
  return true;
}


BibAuthor* Bib::addAuthor(const std::string& lastname, const std::string& firstname)
{
  // authorMap_ indexes the lastname, it points to a list of authors with different
  // firstnames (!!Note: only first letter of the firstname is checked !!)
  BibAuthors* authors = nullptr;
  
  auto it = authorMap_.find(lastname);
  if (it != authorMap_.end()) authors = &it->second;
  else {
    authorMap_[lastname] = BibAuthors();
    it = authorMap_.find(lastname);
    authors = &it->second;
  }
  
  for (auto* a : *authors) {        // just checks first letter of the firstname !!
    if (a->firstname_.empty() || firstname.empty() || a->firstname_[0] == firstname[0])
      return a;
  }
  
  BibAuthor* a = new BibAuthor(lastname, firstname);
  authors->push_back(a);
  return a;
}


BibAuthor* Bib::getAuthor(const std::string& lastname,
                          const std::string& firstname) const {
  const BibAuthors* authors = nullptr;
  auto it = authorMap_.find(lastname);
  if (it == authorMap_.end()) return nullptr;
  else authors = &it->second;
  
  for (auto* a : *authors) {       // just checks first letter of the firstname !!
    if (a->firstname_.empty() || firstname.empty() || a->firstname_[0] == firstname[0])
      return a;
  }
  return nullptr;
}



const BibEntry* Bib::getEntry(const string& key) const {
  auto it = entryMap_.find(key);
  return (it == entryMap_.end()) ? nullptr : it->second;
}


BibEntry* Bib::addEntry(const string& key, const string& bibType, bool isAuxiliary)
{
  string typeName = ccuty::tolower(bibType);             // bibType is lowercased
  BibTypes::ID typeID = BibTypes::addType(typeName);     // adds only if needed
  BibEntry* ref{nullptr};

  switch (typeID) {
    case BibTypes::conference: // BibTex type 'conference' is an alias for 'inproceedings'
      typeName = "inproceedings";
      typeID = BibTypes::inproceedings;
      break;
      
    case BibTypes::knownmedia:
      hasMedias_ = true;
      ref = new BibEntry(ccuty::tolower(key), typeName, typeID, true);
      medias_.push_back(BibEntryNode(*ref));
      return ref;
      
    case BibTypes::style:
      return styleMap_[ccuty::tolower(key)] = new BibEntry(key, typeName, typeID, true);
      
    case BibTypes::authgroup: {
      std::string lowerkey = ccuty::tolower(key);
      auto it = groupMap_.find(lowerkey);
      if (it != groupMap_.end()) {    // dont replace group if already existing
        error("addGroup","group "+key+"already exists - ignoring");
        return nullptr;
      }
      else return groupMap_[lowerkey] = new BibGroup(key, typeName, typeID, true);
    } break;
      
    default:
      break;
  }
  
  auto it = entryMap_.find(key);
  // the first occurence of the entries is kept if it already exists
  if (it != entryMap_.end()) return it->second;
  else {
    ref = new BibEntry(key, typeName, typeID, isAuxiliary);
    allEntries_.push_back(BibEntryNode(*ref));
    entryMap_[key] = ref;
    return ref;
  }
}


void Bib::completeEntry(BibEntry& entry) {
  if (entry.typeID() == BibTypes::knownmedia) {
    // the title must be lowercased in this case
    entry.setParsedField(BibFields::title, ccuty::tolower(entry.field(BibFields::title)));
  }
  else if (entry.typeID() == BibTypes::authgroup) {
    for (auto& a : entry.authors()) {
      BibAuthor* a2 = addAuthor(a->lastname_, a->firstname_);
      BibGroup* gr = dynamic_cast<BibGroup*>(&entry);
      if (!gr) error("completeEntry","wrong Authgroup");
      a2->authgroup_ = gr;
    }
  }
  else {
    // the rank field must not be empty otherwise this entry will appear first
    if (entry.field(BibFields::rank).empty()) entry.setParsedField(BibFields::rank, "?");
    
    // complete doi link if needed
    const string& doi = entry.field(BibFields::doi);
    if (!doi.empty() && doi.compare(0, 4, "http") != 0)
      entry.setParsedField(BibFields::doi, BibSettings::DoiPrefix + doi);
    
    // complete hal link if needed
    const string& hal = entry.field(BibFields::hal);
    if (!hal.empty() && hal.compare(0, 4, "http") != 0)
      entry.setParsedField(BibFields::hal, BibSettings::HalPrefix + hal);
  }
  
  // add bibcite (Bibtex citation field)
  std::ostringstream ss;
  BibWriter writer;
  writer.write(ss, entry, true);
  string s{ss.str()};
  
#ifdef WITH_HYDE
  // needed if hyde is used because {{ and }} have a special meaning
  size_t pos = 0;
  while ((pos = s.find("{{", pos)) != string::npos) s.replace(pos, 2, "{<wbr/>{");
  pos = 0;
  while ((pos = s.find("}}", pos)) != string::npos) s.replace(pos, 2, "}<wbr/>}");
#endif
  entry.setParsedField(BibFields::bibcite, s);

  // call extensions if any
  if (!extensions.empty()) {
    for (auto& it : extensions) it->completeEntry(*this, entry);
  }
}


void Bib::clear() {                 // A COMPLETER !!!!
  // delete current ref list
  if (selectEntries_ != &allEntries_) delete selectEntries_;
  selectEntries_ = nullptr;
  
  // delete refs
  for (auto it : allEntries_) {
    if (it.index_ == 0) delete &it.entry_;
  }
  allEntries_.clear();
  entryMap_.clear();  // points to the same elements as _allEntries!
  
  // delete medias
  for (auto it : medias_) {
    if (it.index_ == 0) delete &it.entry_;
  }
  medias_.clear();

  // delete styles
  for (auto it : styleMap_) delete it.second;
  styleMap_.clear();
}


void Bib::readBib(const string& filename, bool auxiliaryFile) {
  if (filename.empty()) {
    error("read","missing BibTex file");
    return;
  }
  
  cout << "Reading: " << filename << "..." << endl;
  BibReader reader(filename, auxiliaryFile);
  cout <<endl;
  
  selectEntries_ = &allEntries_;
  if (hasMedias_) {      // must search entry medias
    hasMedias_ = false;
    allEntries_.setMediaAndRank(*this);
  }
  
  if (interMode_) writeHtml(true);
}


void Bib::readTemplate(const string& templateFile) {
  if (templateFile.empty() || templateFile == "?") {
    error("templ","Warning, missing HTML template file");
    return;
  }
  templateSearched_ = true;
  
  ifstream f(templateFile);
  if (!f) {
    error("templ","could not open template file '"+templateFile+"'");
    if (templateFile == BibSettings::TemplateFile) return;
    f.open(BibSettings::TemplateFile);
    if (!f.is_open()) {
      error("templ","could not open template file '"+ string(BibSettings::TemplateFile) +"'");
      return;
    }
  }
  htmlBegin_.clear();
  htmlEnd_.clear();
  
  bool inBegin = true;
  string line;
  while (getline(f, line)) {
    ccuty::strtrim(line);
    if (line.find("<!--", 0)!=string::npos && line.find(BibSettings::TemplateMark)!=string::npos) {
      inBegin = false;
      continue;
    }
    if (inBegin) htmlBegin_ += line + '\n';
    else htmlEnd_ += line + '\n';
    
  }
  f.close();
}


void Bib::setHtmlFile(const string& fileName) {
  outHtml_ = ccuty::trimmed(fileName);
  ofstream f(outHtml_);       // should not create file !!!
  if (f) f.close();
  else {
    error("out","can't open '"+fileName+"' for producing HTML file, using '"+BibSettings::OutHtml+"'");
    outHtml_ = BibSettings::OutHtml;   // back to default
    ofstream f2(outHtml_);
    if (f2) f.close();
    else error("out","can't open '"+outHtml_+"' for producing HTML file");
  }
}


void Bib::writeHtml(bool showOnWebBrower) {
  if (!selectEntries_ || selectEntries_->empty()) {
    echo("No entries (-read to read file; -all to select all entries)");
    return;
  }
  if (!templateSearched_) {
    //cerr << "Default template file: "  << endl;
    readTemplate(BibSettings::TemplateFile);
  }
  style_->write(*this, outHtml_, *selectEntries_);
  // show result
  if (showOnWebBrower) ccuty::system(string("open '" + outHtml_ + "'&"));
  echo("Output: " + outHtml_ + "' (" + to_string(selectEntries_->size()) + " entries)");
}


void Bib::setBibFile(const string& fileName) {
  outBib_ = ccuty::trimmed(fileName);
  ofstream f(outBib_);       // should not create file !!!
  if (f) f.close();
  else {
    error("outbib","can't open '"+fileName+"' for producing BibTex file, using '"+BibSettings::OutBib+"'");
    outBib_ = BibSettings::OutBib;
    ofstream f2(outBib_);
    if (f2) f.close();
    else error("outbib","can't open '"+outBib_+"' for producing BibTex file");
  }
}


void Bib::writeBib(bool showOnWebBrower) {
  if (!selectEntries_ || selectEntries_->empty()) {
    echo("No entries (-read to read file; -all to select all entries)");
    return;
  }
  BibWriter writer;
  if (!writer.write(outBib_, *selectEntries_)) {
    echo("!Error: BibTex file '" + outBib_ + "' could not be created");
    return;
  }
  if (showOnWebBrower) ccuty::system(string("open '" + outBib_ + "'&"));
  echo("Output: " + outBib_ + "' (" + to_string(selectEntries_->size()) + " entries)");
}


void Bib::select(const string& pattern) {
  if (pattern.empty() || pattern == "?") {
    echo("Syntax: select field1=\"aaa\" & field2!=\"bbb\" & field3=\"ccc|!ddd\" ...\n"
         "  All conditions betwen & symbols must be met\n"
         "  Examples:\n"
         "   - author=einstein \n"
         "     => entries which author field contains 'einstein' (case is ignored) \n"
         "   - author!=einstein \n"
         "     => author field does NOT contain 'eintein'\n"
         "   - author=einstein|!wilson \n"
         "     => author field contains 'einstein' OR does not contain 'wilson' \n"
         "  Notes:\n"
         "   - any can field existing in the database can be used \n"
         "   - field!="" : returns entries for which this field is not empty \n"
         "   - Field names can be abridged \n"
         );
    // << " * Valid fields are: " << BibDefs::BibFields << endl
    return;
  }
  
  if (allEntries_.empty()) {  // !! allentries ?
    error("select","No entry in the database (-read to read a BibTex file)");
    return;
  }
  
  if (pattern.empty() || pattern == "*") {
    selectEntries_ = &allEntries_;
    if (interMode_) writeHtml(true);
  }
  else {
    BibEntries* entries = selectEntries_->find(pattern, true/*ignorecase*/, true/*abridged*/);
    if (entries) {
      if (selectEntries_ != &allEntries_) delete selectEntries_;
      selectEntries_ = entries;
      if (interMode_) writeHtml(true);
    }
  }
}


void Bib::sort(const string& pattern) {
  if (pattern.empty() || pattern == "?") {
    echo("Syntax: sort field1 & -field2 & ...\n"
         " - -field2 reverses order\n"
         " - Field names can be abridged\n");
    // " * Valid fields are: " << BibDefs::BibFields
    return;
  }
  
  if (!selectEntries_ || selectEntries_->empty()) {
    error("sort","No entries selected (-read to read a BibTex file; -all to select all entries)");
    return;
  }
  
  if (pattern.empty() || pattern == "*") {
    if (interMode_) writeHtml(true);
  }
  else {
    BibEntries* entries = selectEntries_->sort(pattern, true/*abridged*/);
    if (entries) {
      if (selectEntries_ != &allEntries_) delete selectEntries_;
      selectEntries_ = entries;
      if (interMode_) writeHtml(true);
    }
  }
}


void Bib::setFormat(const string& options) {
  if (options.empty() || options == "?") {
    echo("Syntax: format: help not yet available");           // help: to be completed  !!!!
    return;
  }
  std::vector<std::string> tokens;
  ccuty::strsplit(tokens, ccuty::tolower(options), "&");
  
  for (string& tok : tokens) {
    if (tok.empty()) continue;
    string option, arg;
    ccuty::strsplit(option, arg, tok, "=");
  
    switch (ccuty::strid(option)) {
      case ccuty::strid("breaklines"):
        if (arg=="yes" || arg.empty()) breakLines_ = true;
        else breakLines_ = false;
        break;
      case ccuty::strid("images"):
        if (arg=="left" || arg=="yes" || arg.empty()) showImages_ = ShowImages::Left;
        else if (arg=="right") showImages_ = ShowImages::Right;
        else showImages_ = ShowImages::None;
        break;
      case ccuty::strid("links"):
        if (arg=="text" || arg=="yes"  || arg.empty()) showLinks_ = ShowLinks::Text;
        else if (arg=="icon" || arg=="icons") showLinks_ = ShowLinks::Icon;
        else showLinks_ = ShowLinks::None;
        break;
      case ccuty::strid("numbers"):
        if (arg=="yes" || arg.empty()) showNumbers_ = true;
        else showNumbers_ = false;
        break;
      case ccuty::strid("headers"):
        if (arg=="no" || arg.empty()) showHeaders_ = BibFields::none;
        else showHeaders_ = BibFields::addField(arg); // add if not be yet defined;
        break;
      case ccuty::strid("style"):
        setStyle(arg);
        break;
      default:
        error("format","Unknown option: '" + option + "'");
        break;
    }
  }
  if (interMode_ && selectEntries_ && !selectEntries_->empty()) writeHtml(true);
}


void Bib::setStyle(const string& styleKey) {
  if (styleKey.empty()) {
    error("format","missing style name");
    return;
  }
  if (styleKey == "default") {
    delete style_;
    style_ = BibStyle::createStyle(*this, BibSettings::DefaultStyleFormat);
    return;
  }
  auto it = styleMap_.find(styleKey);
  if (it == styleMap_.end() || !it->second) {
    error("format",": style '" + styleKey + "' was not found (style unchanged");
    return;
  }
  
  BibStyle* s = BibStyle::createStyle(*this, *it->second);
  if (!s) error("format"," invalid style '" + styleKey + "' (style unchanged)");
  else {
    delete style_;
    style_ = s;
  }
}


void Bib::showStats(string filename) {
  if (!selectEntries_ || selectEntries_->empty()) {
    error("stat","No entries selected (-read to read file; -all to select all entries)");
    return;
  }
  
  if (filename.empty()) filename = "stats.csv";
  std::string outStats_ = ccuty::trimmed(filename);
  
  bool with_ranks = false;
  //if (arg == "ranks") with_ranks = true;
  
  echo("Writing stats in "+outStats_ + "...\n");
  BibStats stats(outStats_, *selectEntries_, groupMap_, with_ranks);
}


void Bib::checkLinks(const string& fieldName) {
  if (fieldName.empty() || fieldName == "?") {
    error("check","Entry field is missing");
    return;
  }
  BibFields::ID id = BibFields::fieldID(fieldName, true);
  if (id <= BibFields::none) {
    error("check","Unknown entry field: "+ fieldName);
    return;
  }
  
  if (!selectEntries_ || selectEntries_->empty()) {
    error("check","No entries selected (-read to read file; -all to select all entries)");
    return;
  }
  
  cout << "Check '" << fieldName << "' links" << endl;

  for (auto it : *selectEntries_) {
    const string& link = it.field(id);
    cout << " - Entry " << it.bibkey() << ":\t";
    if (link.empty()) cout << "NOLINK" << endl;
    else {
      string com = "curl --output /dev/null --silent --head --fail '"+link+"'";
      if (ccuty::system(com) == 0) cout << "OK" << endl;
      else {
        cout << "INVALID" << endl;
        it.setParsedField(id, "");
      }
    }
  }
  
  if (interMode_) writeHtml(true);
  return;
}

