
// bibhtml: Generates HTML page (with images) from BibTex files
// See: https://www.telecom-paris.fr/~elc/software/
//
// Copyright 2020 Eric Lecolinet (eric.lecolinet@gmail.com)
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

#include <iostream>
#include <sstream>
#include <string>
#include <ctype.h>
#include "ccstring.hpp"
#include "ccpath.hpp"
#include "bib.hpp"
#include "bibentry.hpp"
#include "bibtex.hpp"
using namespace std;

void BibWriter::write(std::ostream& f, const BibEntry& entry, bool inHtmlFile) {
  const string& newline = inHtmlFile ? ",<br>" : ",\n";
  const string& spaces = inHtmlFile ? "&emsp;" : "  ";

  f <<'@' <<  BibTypes::typeName(entry.typeID())
  << "{" << entry.bibkey() << newline;
  
  bool before = false;
  for (auto& it : entry.fields()) {
    if (!inHtmlFile
        || (it.first != BibFields::abstract && it.first != BibFields::keywords
            && it.first != BibFields::annote && it.first != BibFields::note)) {
      if (!it.second.raw.empty()) {
        if (before) f << newline;
        before = true;
        f << spaces << BibFields::fieldName(it.first) << " = " << it.second.raw;
      }
    }
        }
  f << newline << "}\n";
}

bool BibWriter::write(const string& filename, const BibEntries& entries) {
  ofstream f(filename);
  if (!f) return false;
  for (auto& it : entries) {
    write(f, it.entry_, false);
    f << '\n';
  }
  f << endl;
  f.close();
  return true;
}


BibReader::BibReader(const string& filename, bool isAuxiliaryFile) {
  status_ = true;
  entry_ = nullptr;
  isAuxiliary_ = isAuxiliaryFile;
  filename_ = filename;
  lineNo_ = 0;
  
  if (ccuty::extname(filename) != ".bib") {
    cerr << "! Error: Not a BibTex file (no .bib extension): '"+filename+"'" << endl;
    return;
  }
  
  fin_.open(filename);
  if (!fin_) {
    cerr << "! Error: Can't open BibTex file: '"+filename+"'" << endl;
    return;
  }
  
  while (fin_.good()) {
    int c = 0;
    // everything before a @ is a comment
    while (fin_.good() && (c = fin_.get()) != '@') {
      if (c == '\n') ++lineNo_;
    }
    if (fin_.good() && c == '@') readEntry();
  }
  fin_.close();
  /*
  if (!status_)
    cerr << "! Warning: " << status_ << " error(s) found in BibTex file "
    << filename << endl;
   */
}


void BibReader::error(const string& mesg) {
  string s;
  getline(fin_, s);
  status_ = false;
  cerr << "! Error in BibTex file: '" << filename_<< "' near line " << lineNo_
  << ":\n  - " << mesg;
  if (!s.empty()) cerr <<" before: '" << s << "\n" << "'";
  cerr << endl;
}


bool BibReader::readEntry() {
  // read BibTex entry type
  string bibType;
  getline(fin_, bibType, '{');
  ccuty::strtrim(bibType);
  if (bibType.empty()) {
    error("wrong entry: no type name");
    return false;
  }
    
  // read BibTex key
  string bibKey;
  getline(fin_, bibKey, ',');
  ccuty::strtrim(bibKey);
  if (bibKey.empty()) {
    error("wrong entry: no citation key");
    return false;
  }
  entry_ = Bib::instance.addEntry(bibKey, bibType, isAuxiliary_);
  
  // read fields
  while (fin_.good()) {
    char finalChar = readField();
    if (finalChar == 0) return false;
    else if (finalChar == '}') {     // end of entry
      Bib::instance.completeEntry(*entry_);
      return true;
    }
  }
  return false;
}


char BibReader::readField() {
  int c = nextChar();
  if (c=='}') return c;    // no field or field ended by ,}
  else if (c==',') {       // empty field such as: "xxx, ,yyy"
    c = nextChar();
  }
    
  std::string fieldName;
  c = readSlotName(fieldName, c);
  if (c != '=') {
    error("wrong field: expecting =");
    return 0;
  }
  ccuty::strtrim(fieldName);
  if (fieldName.empty()) {
    error("wrong field: no field name");
    return 0;
  }
  
  c = nextChar();
  char finalChar = 0;
  string value;
  
  if (c == '"') finalChar = readQuotedValue(value);
  else if (c == '{') finalChar = readBracedValue(value);
  else finalChar = readUnquotedValue(value, c);
  ccuty::strtrim(value);
  if (finalChar==0) {
    error("wrong entry: expecting , or }");
    return 0;
  }
  
  BibFields::ID id = BibFields::addField(fieldName);
  string parsed;
  if (id == BibFields::author) parseAuthors(value, parsed);
  else parseValue(value, parsed);
  
  entry_->setField(id, value, parsed);
  return finalChar;
}


char BibReader::readSlotName(string& value, char firstChar) {
  value = firstChar;
  int c = 0;
  while (fin_.good() && (::isalnum(c = fin_.get()) || c=='-' || c=='_' || c==':')) value += c;
  if (::isspace(c)) c = nextChar();
  return fin_.good() ? c : 0;
}

char BibReader::readUnquotedValue(string& value, char firstChar) {
  value = firstChar;
  int c = 0;
  while (fin_.good() && (c = fin_.get())!=',' && c!='}') value += c;
  return (fin_.good() && (c==',' || c=='}')) ? c : 0;
}

char BibReader::readQuotedValue(string& value) {
  value.clear();
  int braceCount = 0;
  int c = 0;
  value += '"';
  
  while (fin_.good()) {
    c = fin_.get();
    if (c == '\\') {
      value += c;
      c = fin_.get();
    }
    else if (c == '\n') ++lineNo_;
    else if (c == '{') ++braceCount;
    else if (c == '}') --braceCount;
    else if (c == '"' && braceCount==0) {
      value += c; break;
    }
    value += c;
  }
  c = nextChar(value);
  return (c==',' || c=='}') ? c : 0;
}


char BibReader::readBracedValue(string& value) {
  value.clear();
  int braceCount = 1;
  int c = 0;
  //value += '{';   !!!!
  
  while (fin_.good()) {
    c = fin_.get();
    if (c == '\\') {
      value += c;
      c = fin_.get();
    }
    else if (c == '\n') ++lineNo_;
    else if (c == '{') ++braceCount;
    else if (c == '}') {
      if (--braceCount==0) {
        //value += c; break;  !!!!
        break;
      }
    }
    value += c;
  }
  c = nextChar(value);
  return (c==',' || c=='}') ? c : 0;
}

char BibReader::nextChar(string& value) {
  char c = 0;
  while (fin_.good() && ::isspace(c = fin_.get())) {
    if (c == '\n') ++lineNo_;
    value += c;
  }
  return fin_.good() ? c : 0;
}

char BibReader::nextChar() {
  char c = 0;
  while (fin_.good() && ::isspace(c = fin_.get())) {
    if (c == '\n') ++lineNo_;
  }
  return fin_.good() ? c : 0;
}


static map<char,string> table1 {
  {'n', u8"\n"},
  {'\\', u8"\\"},
  {'&', u8"&"},
  {'o', u8"ø"},
  {'O', u8"Ø"},
  {'l', u8"ł"},
  {'S', u8"§"},
};

static map<string,string> table2 {
  {"'a", u8"á"},
  {"`a", u8"à"},
  {"^a", u8"â"},
  {"~a", u8"ã"},
  {"\"a", u8"ä"},
  {".a", u8"ȧ"},
  {"aa", u8"å"},
  {"'A", u8"Á"},
  {"`A", u8"À"},
  {"^A", u8"Â"},
  {"~A", u8"Ã"},
  {"\"A", u8"Ä"},
  {".A", u8"Ȧ"},
  {"AA", u8"Å"},
  
  {"'e", u8"é"},
  {"`e", u8"è"},
  {"^e", u8"ê"},
  {"~e", u8"ẽ"},
  {"\"e", u8"ë"},
  {".e", u8"ė"},
  {"'E", u8"É"},
  {"`E", u8"È"},
  {"^E", u8"Ê"},
  {"~E", u8"Ẽ"},
  {"\"E", u8"Ë"},
  {".E", u8"Ė"},

  {"'i", u8"í"},
  {"`i", u8"ì"},
  {"^i", u8"î"},
  {"~i", u8""},
  {"\"i", u8"ï"},
  {"'I", u8"Í"},
  {"`I", u8"Ì"},
  {"^I", u8"Î"},
  {"~I", u8"Ĩ"},
  {"\"I", u8"Ï"},
  
  {"'o", u8"ó"},
  {"`o", u8"ò"},
  {"^o", u8"ô"},
  {"~o", u8"õ"},
  {"\"o", u8"ö"},
  {".o", u8"ȯ"},
  {"Ho", u8"ő"},
  {"uo", u8"ǒ"},
  {"'O", u8"Ó"},
  {"`O", u8"Ò"},
  {"^O", u8"Ô"},
  {"~O", u8"Õ"},
  {"\"O", u8"Ö"},
  {".O", u8"Ȯ"},
  {"HO", u8"Ő"},
  {"uO", u8"Ŏ"},
  
  {"'u", u8"ú"},
  {"`u", u8"ù"},
  {"^u", u8"û"},
  {"\"u", u8"ü"},
  {"uu", u8"ů"},
  {"'U", u8"Ú"},
  {"`U", u8"Ù"},
  {"^U", u8"Û"},
  {"\"U", u8"Ü"},
  {"UU", u8"Ů"},
  
  {"cc", u8"ç"},
  {"ss", u8"ß"},
  {"~n", u8"ñ"},
  {"ae", u8"æ"},
  {"oe", u8"œ"},
  {"AE", u8"Æ"},
  {"OE", u8"Œ"},
};

// \textsuperscript{a} and \textcircled{a} are missing
static map<string,string> table3 {
  {"mu", u8"µ"},
  {"degree", u8"°"},
  {"dag", u8"†"},
  {"ddag", u8"‡"},
  {"beta", u8"ß"},
  {"euro", u8"€"},
  {"pounds", u8"£"},
  {"copyright", u8"©"},
  {"guillemotleft", u8"«"},
  {"guillemotright", u8"»"},
  {"textmu", u8"µ"},
  {"backslash", u8"\\"},
  {"textbackslash", u8"\\"},
  {"bullet", u8"•"},
  {"textbullet", u8"•"},
  {"cdot", u8"·"},
  {"textperiodcentered", u8"·"},
  {"textacutedbl", u8"a"},
  {"textdegree", u8"°"},
  {"textquotesingle", u8"'"},
  {"textquoteleft", u8"`"},
  {"textquoteright", u8"'"},
  {"textquotedblleft", u8"``"},
  {"textquotedblright", u8"''"},
  {"textless", u8"<"},
  {"textgreater", u8">"},
  {"mid", u8"|"},
  {"textbar", u8"|"},
  {"textendash", u8"–"},
  {"textemdash", u8"--"},
  {"texttrademark", u8"™"},
  {"textexclamdown", u8"¡"},
  {"textemdash", u8"—"},
  {"textregistered", u8"®"},
  {"textquestiondown", u8"¿"},
  {"textregistered", u8"®"},
  {"textasciiacute", u8"'"},
  {"textasciigrave", u8"`"},
  {"textasciicircum", u8"^"},
  {"textasciitilde", u8"~"},
};

// {\c c} ç

//  \xxx escape, including {\xxx escape called from parseEscape2()
static void parseEscape1(string& res, const string& s, size_t& k, bool from_escape2) {
  size_t len = s.size();

  if (k+4<len && s[k+2]=='{' && s[k+4]=='}') {   // e.g. \'{e}
    char token[] = {s[k+1], s[k+3], 0};
    auto it = table2.find(token);
    if (it != table2.end()) {
      res += it->second;
      k+=5;
      return;
    }
  }
  if (k+2<len) {
    char token[] = {s[k+1], s[k+2], 0};   // e.g. \'e
    auto it = table2.find(token);
    if (it != table2.end()) {
      res += it->second;
      k+=3;
      return;
    }
  }
  if (k+1<len) {
    auto it = table1.find(s[k+1]);    // e.g. \n
    if (it != table1.end()) {
      res += it->second;
      k+=2;
      return;
    }
  }
  
  if (from_escape2) {
    size_t end = 0;
    for (end = k+1; end < len; ++end) {
      if (s[end] == '}') break;
    }
    // if } found substitue if found, always remove this part
    if (end > 0 && end < len) {
      string token = s.substr(k+1, end-k-1);
      auto it = table3.find(token);
      if (it != table3.end()) res += it->second;
      //if (it == table3.end()) cerr << "token: " << token  <<endl;
      k = end+1;
      return;
    }
  }
  
  // default: just keep char after backslash
  if (k+1<len) res += s[k+1];
  k+=2;
}

//  {xxx or {\xxx escape
static void parseEscape2(string& res, const string& s, size_t& k) {
  size_t len = s.size();

  if (k+1 < len && s[k+1]=='\\') {    // e.g. {\'e}
    ++k;
    parseEscape1(res, s, k, true);
    if (k < len && s[k]=='}') k++;  // final }
    return;
  }

  // default: keep {
  res += s[k];
  ++k;
}

void BibReader::parseValue(const string& s, string& res) {
  if (s.empty()) return;
  res.reserve(s.length()+1);
  
  for (size_t k = 0; k < s.size(); ) {      // cf ## et @ non traites !
    if (s[k] == '\n') {
      res += ' ';   // \n replaced by space
      ++k;
    }
    else if (s[k]=='\\') {
      parseEscape1(res, s, k, false);
    }
    else if (s[k] == '{') {
      parseEscape2(res, s, k);
    }
    else if (s[k]=='-' && k+1<s.size() && s[k+1]=='-') {  // "--"
      res += u8"-";  // could be a long--
      k += 2;
    }
    else {   // default: keep char
      res += s[k];
      ++k;
    }
  }
}

// - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - -

static void tokenizeAuthors(const std::string& s, std::vector<std::string>& tokens) {
  size_t len = s.size();
  std::string token;
  size_t k = 0;
  
  while (k < len) {
    char c = s[k];
    
    if (isspace(c) || c == ',') {
      if (!token.empty()) {
        tokens.push_back(token);
        token.clear();
      }
      if (c == ',') {
        tokens.push_back(",");
      }
      k++;
    }
    
    else if (c == '{') {
      k++;
      int parent = 1;
      
      while (k < len) {
        char c2 = s[k];
        
        if (c2 == '{') parent++;   // bug \{
        else if (c2 == '}') {
          parent--;
          if (parent == 0) {k++; break;}
        }
        else token += c2;
        k++;
      }
    }
    
    else {
      token += c;
      k++;
    }
  }
  
  if (!token.empty()) tokens.push_back(token);
}


static void mergeTokens(const std::vector<std::string>& tokens,
                        ssize_t begin, ssize_t end, std::string& name) {
  name.clear();
  if (begin < 0) begin = 0;
  if (end >= tokens.size()) end = tokens.size() -1;
  
  for (size_t k = begin; k <= end; ++k) {
    name += tokens[k];
    if (k < end) name += " ";
  }
}

static void parseAuthor(const std::vector<std::string>& tokens,
                        ssize_t begin, ssize_t end, BibAuthors& authors) {
  std::string firstname, lastname;
  if (begin < 0) begin = 0;
  if (end >= tokens.size()) end = tokens.size() -1;
  
  if (begin > end || end < 0) {
    return;
  }
  else if (begin == end) {
    BibAuthor* a = Bib::instance.addAuthor(tokens[end], "");
    authors.push_back(a);
  }
  else {
    ssize_t comma = -1;
    for (size_t k = begin; k <= end; ++k) {
      if (tokens[k] == ",") comma = k;
    }
    
    if (comma >= 0) {
      mergeTokens(tokens, begin, comma-1, lastname);
      mergeTokens(tokens, comma+1, end, firstname);
    }
    else {
      mergeTokens(tokens, begin, end-1, firstname);
      mergeTokens(tokens, end, end, lastname);
    }
    BibAuthor* a = Bib::instance.addAuthor(lastname, firstname);
    authors.push_back(a);
  }
}


void BibReader::parseAuthors(const std::string& authorsdef, BibAuthors& authors) {
  std::vector<std::string> tokens;
  
  std::string s;
  parseValue(authorsdef, s);
  tokenizeAuthors(s, tokens);
  //for (auto& s : tokens) {cout << "[" << s << "]" << endl;}
  //cout << endl;
  
  ssize_t k = 0, prevk = 0;
  for (auto& tok : tokens) {
    if (tok == "and") {
      parseAuthor(tokens, prevk, k-1, authors);
      prevk = k+1;
    }
    k++;
  }
  if (prevk <= k) parseAuthor(tokens, prevk, k, authors);
}


void BibReader::parseAuthors(const std::string& value, std::string& parsed) {
  parseAuthors(value, entry_->authors_);
  parsed.clear();
  for (auto& a : entry_->authors_) {
    parsed += a->firstname_ + " ";
    parsed += a->lastname_ + ", ";
  }
  if (!parsed.empty()) {parsed.pop_back();parsed.pop_back();}  // remove final ", "
}
